#pragma once 

class CVector : public RwV3d
{
public:
	CVector(void) = default;
	CVector(float x, float y, float z)
	{
		this->x = x;
		this->y = y;
		this->z = z;
	}

  CVector(const RwV3d &v)
  {
    x = v.x;
    y = v.y;
    z = v.z;
  }
  // (0,1,0) means no rotation. So get right vector and its atan
  __always_inline float Heading(void) const { return Atan2(-x, y); }
  __always_inline float Magnitude(void) const { 
#if defined(DC_SH4) && 0  /* Probably not a gain! */
    float w;
    vec3f_length(x, y, z, w);
    return w;
#else
    return Sqrt(x*x + y*y + z*z); 
#endif
}
  __always_inline float MagnitudeSqr(void) const { 
#if defined(DC_SH4) && 0 /* Not a gain, shitty FIPR codegen! */
    return fipr_magnitude_sqr(x, y,z, 0.0f); 
#else
    return x*x + y*y + z*z;
#endif
} 
  __always_inline float Magnitude2D(void) const { return Sqrt(x*x + y*y); }
  float MagnitudeSqr2D(void) const { return x*x + y*y; }
  void Normalise(void);
  
  __always_inline void Normalise2D(void) {
    float sq = MagnitudeSqr2D();
    float invsqrt = RecipSqrt(sq);
    x *= invsqrt;
    y *= invsqrt;
  }

  const CVector &operator+=(CVector const &right) {
    x += right.x;
    y += right.y;
    z += right.z;
    return *this;
  }

  const CVector &operator-=(CVector const &right) {
    x -= right.x;
    y -= right.y;
    z -= right.z;
    return *this;
  }

  const CVector &operator*=(float right) {
    x *= right;
    y *= right;
    z *= right;
    return *this;
  }

  const CVector &operator/=(float right) {
    right = Invert(right);
    x *= right;
    y *= right;
    z *= right;
    return *this;
  }

  CVector operator-() const {
    return CVector(-x, -y, -z);
  }

  const bool operator==(CVector const &right) {
    return x == right.x && y == right.y && z == right.z;
  }

  const bool operator!=(CVector const &right) {
    return x != right.x || y != right.y || z != right.z;
  }

  bool IsZero(void) const { return x == 0.0f && y == 0.0f && z == 0.0f; }
};

inline CVector operator+(const CVector &left, const CVector &right)
{
  return CVector(left.x + right.x, left.y + right.y, left.z + right.z);
}

inline CVector operator-(const CVector &left, const CVector &right)
{
  return CVector(left.x - right.x, left.y - right.y, left.z - right.z);
}

inline CVector operator*(const CVector &left, float right)
{
  return CVector(left.x * right, left.y * right, left.z * right);
}

inline CVector operator*(float left, const CVector &right)
{
  return CVector(left * right.x, left * right.y, left * right.z);
}

inline CVector operator/(const CVector &left, float right)
{
  right = Invert(right);
  return CVector(left.x * right, left.y * right, left.z * right);
}

__always_inline float
DotProduct(const CVector &v1, const CVector &v2)
{
#if defined(DC_SH4) && 0 /* Not a gain, FIPR sucks ass with codegen! */
  return fipr(v1.x, v1.y, v1.z, 0.0f, v2.x, v2.y, v2.z, 0.0f);
#else
  return v1.x*v2.x + v1.y*v2.y + v1.z*v2.z;
#endif
}

CVector CrossProduct(const CVector &v1, const CVector &v2);

inline float
Distance(const CVector &v1, const CVector &v2)
{
  float w;
#if defined(DC_SH4) && 0 /* Probably not a gain! */
  vec3f_distance(v1.x, v1.y, v1.z, v2.x, v2.y, v2.z, w);
  return w;
#else
  return (v2 - v1).Magnitude();
#endif
}

inline float
Distance2D(const CVector &v1, const CVector &v2)
{
  float x = v2.x - v1.x;
  float y = v2.y - v1.y;
  return Sqrt(x*x + y*y);
}

class CMatrix;

CVector Multiply3x3(const CMatrix &mat, const CVector &vec);
CVector Multiply3x3(const CVector &vec, const CMatrix &mat);
CVector operator*(const CMatrix &mat, const CVector &vec);